@@ -15,6 +15,8 @@ from click import ClickException
 from tqdm import tqdm
 
 _uuid = "_b531990e-3187-4b52-be1f-6e4d4d1e40c9"
+_darcs_comment = Path("_darcs", _uuid)
+_env_comment = {"EDITOR": f"mv {_darcs_comment}", "VISUAL": f"mv {_darcs_comment}"}
 _isatty = sys.stdout.isatty()
 _verbose = False
 _devnull = DEVNULL
@@ -48,7 +50,7 @@ class Popen(SPOpen):
             stderr = _devnull
         if not stdin and "input" not in kwargs:
             stdin = _devnull
-        super().__init__(*args, stdin=stdin, **kwargs)
+        super().__init__(*args, stderr=stderr, stdin=stdin, **kwargs)
 
 
 def run(*args, stdout=None, stderr=None, stdin=None, **kwargs):
@@ -154,6 +156,18 @@ def darcs_clone(source, destination):
     run(["darcs", "clone", "--no-working-dir", source, destination], check=True)
 
 
+def git_try_fast_forward(rev, last):
+    """Clone git-repo."""
+    try:
+        run(["git", "merge", "--no-commit", "--ff-only", rev], check=True)
+        return True
+    except CalledProcessError:
+        return False
+    finally:
+        wipe()
+        checkout(last)
+
+
 def git_clone(source, destination):
     """Clone git-repo."""
     run(["git", "clone", source, destination], check=True)
@@ -175,7 +189,7 @@ def get_current_branch():
 def author(rev):
     """Get the author of a commit from git."""
     res = run(
-        ["git", "log", "--pretty=format:'%cN <%cE>'", "--max-count=1", rev],
+        ["git", "log", "--pretty=format:%cN <%cE>", "--max-count=1", rev],
         stdout=PIPE,
         check=True,
     )
@@ -185,17 +199,24 @@ def author(rev):
     return msg
 
 
-def message(rev):
+def onelines(rev, *, last=None):
     """Get the short-message of a commit from git."""
-    res = run(
-        ["git", "log", "--oneline", "--no-decorate", "--max-count=1", rev],
-        stdout=PIPE,
-        check=True,
-    )
+    if last:
+        res = run(
+            ["git", "log", "--oneline", "--no-decorate", f"{last}..{rev}"],
+            stdout=PIPE,
+            check=True,
+        )
+    else:
+        res = run(
+            ["git", "log", "--oneline", "--no-decorate", "--max-count=1", rev],
+            stdout=PIPE,
+            check=True,
+        )
     msg = res.stdout.decode("UTF-8").strip()
     if _verbose:
         print(msg)
-    return msg
+    return msg.splitlines()
 
 
 def get_head():
@@ -211,23 +232,21 @@ def get_head():
     return head
 
 
-def get_rename_diff(rev):
-    """Request the renames of a commit from git."""
-    with Popen(
-        ["git", "show", "--diff-filter=R", rev],
-        stdout=PIPE,
-    ) as res:
-        while line := res.stdout.readline():
-            yield line.decode("UTF-8").strip()
-
-
-def record_all(rev, postfix=""):
+def record_all(rev, *, last=None, postfix=None, comments=None):
     """Record all change onto the darcs-repo."""
-    msg = message(rev)
+    msgs = onelines(rev, last=last)
+    msg = msgs[0]
+    comments = "\n".join(msgs[1:])
     by = author(rev)
     if postfix:
         msg = f"{msg} {postfix}"
+    if comments:
+        msg = f"{msg}\n\n{comments}"
+    with _darcs_comment.open("w", encoding="UTF-8") as f:
+        f.write(msg)
     try:
+        env = dict(os.environ)
+        env.update(_env_comment)
         res = run(
             [
                 "darcs",
@@ -235,13 +254,15 @@ def record_all(rev, postfix=""):
                 "--look-for-adds",
                 "--no-interactive",
                 "--ignore-times",
+                "--edit-long-comment",
                 "--author",
                 by,
                 "--name",
-                msg,
+                "",
             ],
             check=True,
             stdout=PIPE,
+            env=env,
         )
         if _verbose:
             print(res.stdout.decode("UTF-8").strip())
@@ -259,7 +280,6 @@ def get_rev_list(head, base):
             "--reverse",
             "--topo-order",
             "--ancestry-path",
-            "--no-merges",
             f"{base}..{head}",
         ],
         stdout=PIPE,
@@ -293,13 +313,30 @@ class RenameDiffState:
     ORIG_FOUND = 3
 
 
-def get_renames(rev):
+def get_rename_diff(rev, *, last=None):
+    """Request the renames of a commit from git."""
+    assert last != rev
+    if last is None:
+        action = "show"
+        range = rev
+    else:
+        action = "diff"
+        range = f"{last}..{rev}"
+    with Popen(
+        ["git", action, "--diff-filter=R", range],
+        stdout=PIPE,
+    ) as res:
+        while line := res.stdout.readline():
+            yield line.decode("UTF-8").strip()
+
+
+def get_renames(rev, *, last=None):
     """Parse the renames from a git-rename-diff."""
     s = RenameDiffState
     state = s.INIT
     orig = ""
     new = ""
-    for line in get_rename_diff(rev):
+    for line in get_rename_diff(rev, last=last):
         if state == s.INIT:
             if line.startswith("diff --git"):
                 state = s.IN_DIFF
@@ -317,26 +354,26 @@ def get_renames(rev):
                 state = s.IN_DIFF
 
 
-def record_revision(rev):
+def record_revision(rev, *, last=None):
     """Record a revision, pre-record moves if there are any."""
     iters = 0
     count = 0
     renames = 0
-    for _ in get_renames(rev):
+    for _ in get_renames(rev, last=last):
         renames += 1
 
     if renames:
         with tqdm(desc="moves", total=renames, leave=False, disable=_disable) as pbar:
-            for orig, new in get_renames(rev):
+            for orig, new in get_renames(rev, last=last):
                 move(orig, new)
                 iters += 1
                 if iters % 50 == 0:
-                    record_all(rev, f"move({count:03d})")
+                    record_all(rev, postfix=f"move({count:03d})")
                     count += 1
                 pbar.update()
     wipe()
     checkout(rev)
-    record_all(rev)
+    record_all(rev, last=last)
 
 
 def get_lastest_rev():
@@ -381,21 +418,25 @@ def less_boring():
     disable.rename(bfile)
 
 
-def transfer(gen, count):
+def transfer(gen, count, *, last=None):
     """Transfer the git-commits to darcs."""
     try:
-        last = None
         with tqdm(desc="commits", total=count, disable=_disable) as pbar:
-            iters = 0
+            records = 0
             for rev in gen:
-                record_revision(rev)
+                if git_try_fast_forward(rev, last):
+                    record_revision(rev, last=last)
+                    last = rev
+                    recorded = True
+                    records += 1
+                    if records % 100 == 0:
+                        checkpoint(last)
                 pbar.update()
-                last = rev
-                iters += 1
-                if iters % 100 == 0:
-                    checkpoint(last)
                 if _shutdown:
                     sys.exit(0)
+            if not recorded:
+                record_revision(rev, last=last)
+                last = rev
     except Exception:
         print(f"Failed on revision {last}")
         raise
@@ -418,7 +459,7 @@ def import_range(rbase):
     checkout(rbase)
     with less_boring():
         record_all(rbase)
-        transfer(gen, count)
+        transfer(gen, count, last=rbase)
 
 
 def import_one():
@@ -426,7 +467,8 @@ def import_one():
     head = get_head()
     wipe()
     checkout(head)
-    transfer([head], 1)
+    record_all(head)
+    checkpoint(head)
 
 
 def fix_pwd():
@@ -500,8 +542,8 @@ def clone(source, destination, verbose):
 def update(verbose, base, warn, shallow):
     """Incremental import of git into darcs.
 
-    By default it imports a shallow copy (the current commit). Use `--no-shallow` to
-    import the complete history.
+    By default it imports a shallow copy (the current commit). Use `--no-shallow`
+    to import the complete history.
     """
     setup(warn, verbose=verbose)
     if not Path(".git").exists():
